#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define LISTENQ 64

#include <sys/types.h>
#include <sys/stat.h>
#ifdef WIN32
#include <windows.h>
#include <winsock2.h>
static int wsock_up = 0;
#define MAX_RECV 65536
#else
#define MAX_RECV (512*1025)
#define closesocket(C) close(C)
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
/* and we enable IPv6 if we see it */
#ifdef AF_INET6
#define USE_IPV6 1
#endif
#endif
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>

static int verb;

/* sbthread */
#ifdef WIN32
#include <windows.h>

#define decl_sbthread DWORD WINAPI
#define sbthread_result(A) (DWORD)(A)
static int sbthread_create(LPTHREAD_START_ROUTINE sa, void *par) {
    DWORD tid;
    return CreateThread(0,0,sa,par,0,&tid);
}

/* FIXME: add mutex support here */

#else
#include <pthread.h>

#define decl_sbthread void *
#define sbthread_result(A) (void *)(A)
static int sbthread_create(void * (thr)(void *), void *par) {
    pthread_t Thread;
    pthread_attr_t ThreadAttr;

    pthread_attr_init(&ThreadAttr);
    pthread_attr_setdetachstate(&ThreadAttr,PTHREAD_CREATE_DETACHED);
    return pthread_create(&Thread,&ThreadAttr,*thr,par);
}

#define sbthread_mutex pthread_mutex_t

sbthread_mutex *sbthread_create_mutex() {
    pthread_mutex_t lm = PTHREAD_MUTEX_INITIALIZER;
    pthread_mutex_t *m = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
    memcpy(m, &lm, sizeof(pthread_mutex_t));
    return m;
}

#define sbthread_lock_mutex(M) pthread_mutex_lock(M);
#define sbthread_unlock_mutex(M) pthread_mutex_unlock(M);
void sbthread_destroy_mutex(sbthread_mutex *m) {
    pthread_mutex_destroy(m);
    free(m);
}

#endif

#include "RSprotocol.h"

/* -- provide ptoi() according to the machine endianness -- */
#if !defined __LITTLE_ENDIAN__ && (defined __x86_64__ || defined __i386__) /* hack for common x86 archs and systems that fail to set __LITTLE_ENDIAN__ */
#define __LITTLE_ENDIAN__ 1
#endif
/* reasonble platforms set __*_ENDIAN__ accordingly */
#ifdef __LITTLE_ENDIAN__
#define ptoi(X) X
#else
#define U_I(X) ((unsigned int)(b[X]))
static int swap_i(int x) {
    unsigned char *b = (unsigned char*) &x;
    return (int) (U_I(0) | (U_I(1) << 8) | (U_I(2) << 16) | (U_I(3) << 24));
}
#ifdef __BIG_ENDIAN__
#define ptoi(X) swap_i(X)
#else
static const int end_const = 1;
static int ptoi(int x) {
    return (((const char*)&end_const)[0] == 1) ? x : swap_i(x);
}
#endif
#endif
/* -- end of ptoi() -- */

#define ALLOW_TARGET_LIST 0x001 /* allow CMD_PROXY_TARGET but only if it matches one entry in the ring */
#define ALLOW_TARGET_ANY  0x002 /* allow any CMD_PROXY_TARGET (overrides the above) */

typedef struct host_ring {
    char *name;
    int port;
    int max_con, cur_con;
    struct host_ring *next;
} host_ring_t;

typedef struct proxy {
    int ser, cli, active, err, port, tracking;
    int allow;
    char *host;
    char *buf;
    host_ring_t *host_list, *current_host;
    int buf_size;
} proxy_t;

static sbthread_mutex *ring_mutex;

static int run_proxy(proxy_t *p) {
    int ser = -1, cli = p->cli, maxfd = cli, stage = 0, family, connected = 0, port = p->port;
    char *host = p->host;
    char *buf;

    if (!p->buf_size)
	p->buf_size = 128*1024;
    if (!p->buf)
	p->buf = (char*) malloc(p->buf_size);
    if (!(buf = p->buf)) {
	fprintf(stderr, "ERROR: cannot start proxy thread - unable to allocate buffer memory\n");
	return 1;
    }

    if (!p->port && !p->host) { /* non-transparent proxy mode */
	static const char *proxy_IDstring="RSpx0103QAP1\r\n\r\n--------------\r\n";
	int n;
#ifdef TCP_NODELAY
	{
	    int opt = 1;
	    setsockopt(cli, IPPROTO_TCP, TCP_NODELAY, (const char*) &opt, sizeof(opt));
	}
#endif
#ifdef PROXY_DEBUG
	printf("proxy %p: no host specified entring non-transparent mode\n", p);
#endif
	n = send(cli, proxy_IDstring, 32, 0);
	if (n != 32) {
	    fprintf(stderr, "WARN: unable to send RSpx ID to client\n");
	    return 30;
	}
	/* waiting for CMD_PROXY_* */
	while (p->active) {
	    struct timeval timv;
	    int s;
	    fd_set readfds;
	    /* 250ms reaction time shold be enough */
	    timv.tv_sec = 0; timv.tv_usec = 250000;
	    FD_ZERO(&readfds);
	    FD_SET(cli, &readfds);
	    s = select(cli + 1, &readfds, 0, 0, &timv);
	    if (s > 0) {
		struct phdr hdr;
		long len;
#ifdef PROXY_DEBUG
		printf("proxy %p: incoming client proxy data\n", p);
#endif
		int n = recv(cli, &hdr, sizeof(struct phdr), 0), cmd;
		if (n == 0) { /* clean exit, connection closed */
		    return 0;
		}
		if (n != sizeof(struct phdr)) {
		    fprintf(stderr, "WARN: client message recv failed n = %d, closing\n", n);
		    return 10;
		}
		cmd = ptoi(hdr.cmd);
		if (CMD_CLASS(cmd) == CMD_CLASS_PROXY) { /* proxy-class commands are not passed on but processed by us */
		    int ptr = 0;
		    len = (unsigned int) ptoi(hdr.len);
#ifdef __LP64__
		    if (hdr.res) {
			long hil = ptoi(hdr.res);
			hil <<= 32;
			len |= hil;
		    }
#endif
		    if (len >= p->buf_size) { /* proxy command too, big - drop and close; we want to load the proxy payload into one buffer for processing */
			fprintf(stderr, "WARN: proxy command with too large payload (%ld), assuming fraud, closing\n", len);
			return 14;
		    }
		    while (len > 0) {
			int sl = len;
			n = recv(cli, p->buf + ptr, sl, 0);
			if (n != sl) {
			    fprintf(stderr, "WARN: client recv required %d bytes, got %d (with %ld to go), closing\n", sl, n, len);
			    return 12;
			}
			ptr += n;
			p->buf[ptr] = 0; /* guarantee a trailing NUL */
			len -= n;
		    }
		    if (cmd == CMD_PROXY_TARGET) {
			char *buf = p->buf;
			char *t_host = buf, *c;
			int t_port = 6311;
			if (*buf == '[' && (c = strchr(buf + 1, ']'))) { /* literal IPv6 address */
			    *(c++) = 0;
			    t_host = buf + 1;
			    if (*c == ':')
				t_port = atoi(c + 1);
			} else if ((c = strchr(buf, ':'))) { /* has port */
			    *(c++) = 0;
			    t_port = atoi(c);
			}
			if (verb) printf("INFO: request to connect to target %s, port %d\n", t_host, t_port);
			if (p->allow & ALLOW_TARGET_ANY) {
			    port = t_port;
			    host = t_host;
			    break; /* continue to transparent mode */
			}
			if (p->allow & ALLOW_TARGET_LIST) {
			    host_ring_t *ring = p->host_list;
			    host_ring_t *root = ring;
			    int allow = 0;
			    while (ring) {
				if (ring->port == t_port && ring->name && !strcmp(ring->name, t_host)) {
				    p->current_host = ring;
				    host = t_host;
				    port = t_port;
				    allow = 1;
				    break;
				}
				ring = ring->next;
				/* support both NULL-terminated and infinite lists */
				if (ring == root) break;
			    }
			    if (allow)
				break;
			    fprintf(stderr, "WARN: CMD_PROXY_TARGET host %s, port %d is not on the allowed list, refusing\n", t_host, t_port);
			    return 32;
			}
			fprintf(stderr, "WARN: CMD_PROXY_TARGET is disallowed by configuration, closing\n");
			return 33;
		    } else if (cmd == CMD_PROXY_GET_SLOT) {
			/* let's see if we have a reason to ask */
			if (p->tracking == 0 && p->current_host && p->current_host->max_con > 0) {
			    sbthread_lock_mutex(ring_mutex);
			    if (p->current_host->cur_con < p->current_host->max_con) { /* can we go? */
				p->tracking = 1;
				p->current_host->cur_con++;
			    }
			    sbthread_unlock_mutex(ring_mutex);
			    if (!p->tracking) { /* not tracking = not available */
				if (verb) printf("INFO: client issued CMD_PROXY_GET_SLOT and the host is full, closing\n");
				return 50;
			    }
			}
			/* either no limit or lock acquired => everything is a go */
		    } else
			fprintf(stderr, "WARN: unsupported proxy command (%x), ignoring\n", cmd);
		} else { /* non-proxy command */
		    fprintf(stderr, "WARN: non-proxy command (0x%x) in proxy mode, closing connection\n", cmd);
		    return 31;
		}
	    }
	}
    }

    if (verb && p->current_host)
	printf("INFO: current_host set to host %s, port %d, max=%d, current=%d\n",
	       p->current_host->name, p->current_host->port, p->current_host->max_con, p->current_host->cur_con);
    
    /* check whether we need to track and wait */
    if (p->tracking == 0 && p->current_host && p->current_host->max_con > 0) {
	/* tracking necessary - the host has connection limit */
	while (1) {
	    sbthread_lock_mutex(ring_mutex);
	    if (p->current_host->cur_con < p->current_host->max_con) { /* can we go? */
		p->tracking = 1;
		p->current_host->cur_con++;
		sbthread_unlock_mutex(ring_mutex);
		break;
	    }
	    sbthread_unlock_mutex(ring_mutex);
	    /* to prevent spinning, we use void select to timeout after 100ms */
	    {
		struct timeval tv;
		tv.tv_sec = 0;
		tv.tv_usec = 100000;
		select(0, NULL, NULL, NULL, &tv);
	    }
	}
    }

    p->ser = -1;
#ifdef WIN32
    family = AF_INET;
#else
    family = port ? AF_INET : AF_LOCAL;
#endif
#ifdef USE_IPV6
    /* we use getaddrinfo to have the system figure the family and address for us */
    /* FIXME: is there any reason we don't use that in general? Do all systems support this? */
    if (host && family == AF_INET) {
        struct addrinfo hints, *ail = 0, *ai;
        char port_s[8];
        snprintf(port_s, sizeof(port_s), "%d", port);
        memset(&hints, 0, sizeof(hints));
        hints.ai_family = PF_UNSPEC;
        hints.ai_socktype = SOCK_STREAM;
        if (getaddrinfo(host, port_s, &hints, &ail) == 0) {
            for (ai = ail; ai; ai = ai->ai_next)
                if (ai->ai_family == AF_INET || ai->ai_family == AF_INET6) {
                    ser = socket(ai->ai_family, SOCK_STREAM, ai->ai_protocol);
                    if (ser != -1) {
                        if (connect(ser, ai->ai_addr, ai->ai_addrlen) == 0)
                            break;
                        /* didn't work - try another address (if ther are any) */
                        closesocket(ser);
                        ser = -1;
                    }
                }
            if (ail)
                freeaddrinfo(ail);
        }
        if (ser != -1) /* the socket will be valid only if connect() succeeded */
            connected = 1;
    } else
#endif
        ser = socket(family, SOCK_STREAM, 0);

    if (ser != -1 && !connected) {
        if (family == AF_INET) {
            struct sockaddr_in sin;
            struct hostent *haddr;
            sin.sin_family = AF_INET;
            sin.sin_port = htons(port);
            if (host) {
#ifdef WIN32
                int al = sizeof(sin);
                if (WSAStringToAddress((LPSTR)host, sin.sin_family, 0, (struct sockaddr*)&sin, &al) != 0) {
                    if (!(haddr = gethostbyname(host))) { /* DNS failed, */
                        closesocket(ser);
                        ser = -1;
                    }
                    sin.sin_addr.s_addr = *((uint32_t*) haddr->h_addr); /* pick first address */
		}
                /* for some reason Windows trashes the structure so we need to fill it again */
                sin.sin_family = AF_INET;
                sin.sin_port = htons(port);
#else
                if (inet_pton(sin.sin_family, host, &sin.sin_addr) != 1) { /* invalid, try DNS */
                    if (!(haddr = gethostbyname(host))) { /* DNS failed, */
                        closesocket(ser);
                        ser = -1;
                    }
                    sin.sin_addr.s_addr = *((uint32_t*) haddr->h_addr); /* pick first address */
                }
#endif
            } else
                sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
            if (ser != -1 && connect(ser, (struct sockaddr*)&sin, sizeof(sin))) {
                closesocket(ser);
                ser = -1;
            }
        } else {
#ifndef WIN32
            struct sockaddr_un sau;
            memset(&sau, 0, sizeof(sau));
            sau.sun_family = AF_LOCAL;
            if (strlen(host) + 1 > sizeof(sau.sun_path)) {
                closesocket(ser);
                ser = -1;
            } else
                strcpy(sau.sun_path, host);
            if (ser != -1 && connect(ser, (struct sockaddr*)&sau, sizeof(sau))) {
                closesocket(ser);
                ser = -1;
            }
#else /* this should never happen */
            ser = -1; 
#endif
        }
    }
    p->ser = ser;
    if (ser == -1) {
	fprintf(stderr, "ERROR: connection to %s:%d failed\n", host, port);
	return 20;
    }
#ifdef PROXY_DEBUG
    printf("proxy %p: connected (ser = %d, cli = %d)\n", p, ser, cli);
#endif
    if (ser > cli) maxfd = ser;
    maxfd++;
    
#ifdef TCP_NODELAY
    {
	int opt = 1;
	setsockopt(ser, IPPROTO_TCP, TCP_NODELAY, (const char*) &opt, sizeof(opt));
	setsockopt(cli, IPPROTO_TCP, TCP_NODELAY, (const char*) &opt, sizeof(opt));
    }
#endif

    while (p->active) {
	struct timeval timv;
	int s;
	fd_set readfds;
	/* 250ms reaction time shold be enough */
	/* timv.tv_sec = 0; timv.tv_usec = 250000;*/
	timv.tv_sec = 5; timv.tv_usec = 0;
	FD_ZERO(&readfds);
	FD_SET(ser, &readfds);
	if (stage > 0) FD_SET(cli, &readfds);
	s = select(maxfd, &readfds, 0, 0, &timv);
#ifdef PROXY_DEBUG
	printf("proxy %p: select = %d\n", p, s);
#endif
	if (s > 0) {
	    /* server has data to send */
	    if (FD_ISSET(ser, &readfds)) {
#ifdef PROXY_DEBUG
		printf("proxy %p: incoming server data, stage = %d\n", p, stage);
#endif
		/* read ID */
		if (stage == 0) {
		    int n = recv(ser, buf, 32, 0);
		    if (n < 32) {
			fprintf(stderr, "WARN: cannot obtain ID from server (recv = %d), closing\n", n);
			return 2;
		    }
		    n = send(cli, buf, 32, 0);
		    if (n != 32) {
			fprintf(stderr, "WARN: sending ID to client failed (send = %d), closing\n", n);
			return 3;
		    }
		    stage++;
		} else { /* message passing */
		    struct phdr hdr;
		    long len;
		    int n = recv(ser, &hdr, sizeof(struct phdr), 0);
		    if (n != sizeof(struct phdr)) {
			fprintf(stderr, "WARN: server message recv failed n = %d, closing\n", n);
			return 4;
		    }
		    n = send(cli, &hdr, sizeof(struct phdr), 0);
		    if (n != sizeof(struct phdr)) {
			fprintf(stderr, "WARN: send message header to client failed (send = %d), closing\n", n);
			return 5;
		    }
		    len = (unsigned int) ptoi(hdr.len);
#ifdef __LP64__
		    if (hdr.res) {
			long hil = ptoi(hdr.res);
			hil <<= 32;
			len |= hil;
		    }
#endif
		    while (len > 0) {
			int sl = (len > p->buf_size) ? p->buf_size : len;
			n = recv(ser, p->buf, sl, 0);
			if (n != sl) {
			    fprintf(stderr, "WARN: srv recv required %d bytes, got %d (with %ld to go), closing\n", sl, n, len);
			    return 6;
			}
			n = send(cli, p->buf, sl, 0);
			if (n != sl) {
			    fprintf(stderr, "WARN: srv->cli attempt to send %d bytes, result %d (with %ld to go), closing\n", sl, n, len);
			    return 7;
			}
			len -= n;
		    }
		}
	    }
	    /* client has data to send */
	    if (FD_ISSET(cli, &readfds)) {
		struct phdr hdr;
		long len;
#ifdef PROXY_DEBUG
		printf("proxy %p: incoming client data\n", p);
#endif
		int n = recv(cli, &hdr, sizeof(struct phdr), 0), cmd;
		if (n == 0) { /* clean exit, connection closed */
		    return 0;
		}
		if (n != sizeof(struct phdr)) {
		    fprintf(stderr, "WARN: client message recv failed n = %d, closing\n", n);
		    return 10;
		}
		cmd = ptoi(hdr.cmd);
		if (CMD_CLASS(cmd) == CMD_CLASS_PROXY) { /* proxy-class commands are not passed on but processed by us */
		    int ptr = 0;
		    len = (unsigned int) ptoi(hdr.len);
#ifdef __LP64__
		    if (hdr.res) {
			long hil = ptoi(hdr.res);
			hil <<= 32;
			len |= hil;
		    }
#endif
		    if (len >= p->buf_size) { /* proxy command too, big - drop and close; we want to load the proxy payload into one buffer for processing */
			fprintf(stderr, "WARN: proxy command with too large payload (%ld), assuming fraud, closing\n", len);
			return 14;
		    }
		    while (len > 0) {
			int sl = len;
			n = recv(cli, p->buf + ptr, sl, 0);
			if (n != sl) {
			    fprintf(stderr, "WARN: client recv required %d bytes, got %d (with %ld to go), closing\n", sl, n, len);
			    return 12;
			}
			ptr += n;
			p->buf[ptr] = 0; /* guarantee a trailing NUL */
			len -= n;
		    }
		    if (cmd == CMD_PROXY_TARGET) {
			char *buf = p->buf;
			char *t_host = buf, *c;
			int t_port = 6311;
			if (*buf == '[' && (c = strchr(buf + 1, ']'))) { /* literal IPv6 address */
			    *(c++) = 0;
			    t_host = buf + 1;
			    if (*c == ':')
				t_port = atoi(c + 1);
			} else if ((c = strchr(buf, ':'))) { /* has port */
			    *(c++) = 0;
			    t_port = atoi(c);
			}
			printf("INFO: request to connect to target %s, port %d\n", t_host, t_port);
		    } else
			fprintf(stderr, "WARN: unsupported proxy command (%x), ignoring\n", cmd);
		} else {
		    n = send(ser, &hdr, sizeof(struct phdr), 0);
		    if (n != sizeof(struct phdr)) {
			fprintf(stderr, "WARN: send message header to server failed (send = %d), closing\n", n);
			return 11;
		    }
		    len = (unsigned int) ptoi(hdr.len);
#ifdef __LP64__
		    if (hdr.res) {
			long hil = ptoi(hdr.res);
			hil <<= 32;
			len |= hil;
		    }
#endif
		    while (len > 0) {
			int sl = (len > p->buf_size) ? p->buf_size : len;
			n = recv(cli, p->buf, sl, 0);
			if (n != sl) {
			    fprintf(stderr, "WARN: client recv required %d bytes, got %d (with %ld to go), closing\n", sl, n, len);
			    return 12;
			}
			n = send(ser, p->buf, sl, 0);
			if (n != sl) {
			    fprintf(stderr, "WARN: srv<-cli attempt to send %d bytes, result %d (with %ld to go), closing\n", sl, n, len);
			    return 13;
			}
			len -= n;
		    }
		}	
	    }
	}
    }
    return 0;
}

static decl_sbthread proxy_thread(void *par) {
    proxy_t *p = (proxy_t*) par;
    p->err = -1;
#ifdef PROXY_DEBUG
    printf("proxy %p: thread started, run_proxy()\n", p);
#endif
    p->err = run_proxy(p);
    if (verb)
	printf("INFO: run_proxy(%p) returned with %d\n", p, p->err);
#if 0 /* do not release buffers so we an re-use them (maybe some house-cleaning later?) */
    if (p->buf) {
	free(p->buf);
	p->buf = 0;
    }
#endif
    if (p->ser != -1) closesocket(p->ser);
    if (p->cli != -1) closesocket(p->cli);
    p->active = 0;
    if (p->tracking) {
	sbthread_lock_mutex(ring_mutex);
	if (p->current_host)
	    p->current_host->cur_con--;
	p->tracking = 0;
	sbthread_unlock_mutex(ring_mutex);
    }
    return 0;
}

#define MAX_THREADS 2048

static proxy_t proxy[MAX_THREADS];
static int proxies;

/* -- this is a slimmed-down version from Rserve (sisocks + RSserver) --*/
#ifdef WIN32
static int initsocks(void)
{
  WSADATA dt;
  /* initialize WinSock 1.1 */
  return (WSAStartup(0x0101,&dt))?-1:0;
}
#else
#define initsocks()
#endif

/* definition of a server */
typedef struct server {
    int ss;               /* server socket */
    int unix_socket;      /* 0 = TCP/IP, 1 = unix socket */
    int flags;            /* optional server-specific flags */
} server_t;

#define SRV_IPV6  0x1000 /* use IPv6 */
#define SRV_LOCAL 0x4000 /* bind to local loopback interface only */

static server_t *create_server(int port, const char *localSocketName, int localSocketMode, int flags) {
    server_t *srv;
    struct sockaddr_in ssa;
    int reuse, ss, r;
#ifdef HAVE_IPV6
    struct sockaddr_in6 ssa6;
#endif
#ifndef WIN32
    struct sockaddr_un lusa;
#endif

    initsocks();
    if (localSocketName) {
#ifdef WIN32
	fprintf(stderr,"Local sockets are not supported on non-unix systems.\n");
	return 0;
#else
	ss = socket(AF_LOCAL, SOCK_STREAM, 0);
	memset(&lusa, 0, sizeof(lusa));
	lusa.sun_family = AF_LOCAL;
	if (strlen(localSocketName) > sizeof(lusa.sun_path) - 2) {
	    fprintf(stderr,"Local socket name is too long for this system.\n");
	    return 0;
	}
	strcpy(lusa.sun_path, localSocketName);
	remove(localSocketName); /* remove existing if possible */
#endif
    } else
#ifdef HAVE_IPV6
	ss = socket((flags & SRV_IPV6) ? AF_INET6 : AF_INET, SOCK_STREAM, 0);
#else
        ss = socket(AF_INET, SOCK_STREAM, 0);
#endif
    if (ss == -1) {
	perror("ERROR: cannot create socket for the server");
	return 0;
    }
    srv = (server_t*) calloc(1, sizeof(server_t));
    if (!srv) {
	fprintf(stderr, "ERROR: cannot allocate memory for server structure\n");
	closesocket(ss);
	return 0;
    }

    srv->ss = ss;
    srv->unix_socket = localSocketName ? 1 : 0;
    srv->flags = flags;
    
#ifndef WIN32
    reuse = 1; /* enable socket address reusage */
    setsockopt(ss, SOL_SOCKET, SO_REUSEADDR, (const char*)&reuse, sizeof(reuse));
    
    if (localSocketName) {
	r = bind(ss, (struct sockaddr*) &lusa, sizeof(lusa));
	if (localSocketMode)
	    chmod(localSocketName, localSocketMode);
    } else {
#endif
#ifdef HAVE_IPV6
	if (flags & SRV_IPV6) {
	    memset(&ssa6, 0, sizeof(ssa6));
	    ssa6.sin6_family = AF_INET6;
	    ssa6.sin6_port = htons(port);
	    ssa6.sin6_addr = (flags & SRV_LOCAL) ? in6addr_loopback : in6addr_any;
	    r = bind(ss, (struct sockaddr*) &ssa6, sizeof(ssa6));
	} else {
#endif
	    memset(&ssa, 0, sizeof(ssa));
	    ssa.sin_family = AF_INET;
	    ssa.sin_port = htons(port);
	    ssa.sin_addr.s_addr = htonl((flags & SRV_LOCAL) ? INADDR_LOOPBACK : INADDR_ANY);
	    r = bind(ss, (struct sockaddr*) &ssa, sizeof(ssa));
#ifdef HAVE_IPV6
	} /* if (flags & SRV_IPV6) else */
#endif
#ifndef WIN32
    } /* if (localSocketName) else */
#endif
    if (r) {
	perror("ERROR: bind failed");
	closesocket(ss);
	free(srv);
	return 0;
    }

    r = listen(ss, LISTENQ);
    if (r) {
	perror("ERROR: listen failed");
	closesocket(ss);
	free(srv);
	return 0;
    }
    
    return srv;
}

static int active = 1;

int main(int ac, char **av) {
    host_ring_t *root = 0, *ring = 0;
    /* bind, listen, initialize proxy_t, spawn thread */
    server_t *srv;
    int i = 0, local = 0, ipv6 = 0, port = 6311, allow_any = 0, RSPX_mode = 0, gmc = 0;

    while (++i < ac)
	if (av[i][0] == '-')
	    switch (av[i][1]) {
	    case 'l': local = 1; break;
	    case '6': ipv6 = 1; break;
	    case 'v': verb++; break;
	    case 'a': allow_any = 1; break;
	    case 'x': RSPX_mode = 1; break;
	    case 'm': if (++i < ac) gmc = atoi(av[i]); break;
	    case 'p': if (++i < ac) port = atoi(av[i]); break;
	    case 'h': printf("\n Usage: %s [-v] [-p <port>] [-m <max.con>] [-l] [-6] <host1[:port]> [<host2[:port]> [...]]\n\n -p = set port to listen on (default 6311)\n -l = listen on loopback interface only\n -6 = use IPv6\n -m = limit the max. number of connections per host\n      (affects only hosts listed *after* this option)\n -a = allow any target (valid with -x only, no hosts list needed)\n -x = use non-transparent RSpx mode requiring clients\n      to send CMD_PROXY_TARGET first (requires clients capable of dealing\n      with proxies).\n\n NOTE: literal IPv6 addresses must be quoted in []\n\n", av[0]); return 0;
	    }
	else {
	    int port = 6311;
	    char *dc, *hc = av[i];
	    host_ring_t *h = (host_ring_t*) calloc(1, sizeof(host_ring_t));
	    if (hc[0] == '[' && (dc = strchr(hc, ']'))) { /* literal IPv6 address */
		hc++;
		*(dc++) = 0;
	    } else dc = strchr(hc, ':');
	    if (dc && *dc == ':') {
		*(dc++) = 0;
		port = atoi(dc);
	    } else dc = hc;
	    h->name = strdup(hc);
	    h->port = port;
	    h->max_con = gmc;
	    if (root) {
		ring->next = h;
		h->next = root;
		ring = h;
	    } else
		h->next = ring = root = h;
	}
    
    if (!root && !(allow_any && RSPX_mode)) {
	fprintf(stderr, "ERROR: no target hosts specified, but required (see %s -h if in doubt)\n", av[0]);
	return 0;
    }

    ring_mutex = sbthread_create_mutex();
    srv = create_server(port, 0, 0, (local ? SRV_LOCAL : 0) | (ipv6 ? SRV_IPV6 : 0));
    if (srv) {
	while (active) {
	    struct sockaddr_in lsa;
	    socklen_t al = sizeof(lsa);
	    int s = accept(srv->ss, (struct sockaddr*)&lsa, &al), i = 0;
	    if (s == -1) break;
	    while (i < proxies && proxy[i].active) i++;
	    if (i >= MAX_THREADS) {
		fprintf(stderr, "ERROR: reached max. number of threads, dropping\n");
		close(s);
	    } else {
		proxy[i].cli = s;
		proxy[i].err = 0;
		proxy[i].tracking = 0;
		if (RSPX_mode) {
		    proxy[i].port = 0;
		    proxy[i].host = 0;
		    proxy[i].current_host = 0;
		    if (allow_any)
			proxy[i].allow = ALLOW_TARGET_ANY;
		    else {
			proxy[i].allow = ALLOW_TARGET_LIST;
			proxy[i].host_list = root;
		    }
		} else {
		    proxy[i].port = ring->port;
		    proxy[i].host = ring->name;
		    proxy[i].current_host = ring;
		    proxy[i].allow = 0;
		    ring = ring->next;
		}
		proxy[i].active = 1;
		if (i >= proxies) proxies = i + 1;
		if (verb) printf("INFO: connection accepted, spawning new thread on slot %d (%p) -> %s:%d\n", i, proxy + i, proxy[i].host ? proxy[i].host : "<host/port via RSpx>", proxy[i].port);
		sbthread_create(proxy_thread, proxy + i);
	    }
	}
    }
    return 0;
}

#ifdef R_PACKAGE

#include <Rinternals.h>

SEXP proxy() {
}

#endif
